import secrets
import string
from typing import Tuple, Optional, Dict, Any, List

import click
from rossum import option, argument
from rossum.lib import QUEUES
from rossum.lib.api_client import RossumClient
from tabulate import tabulate


@click.group("connector")
def cli() -> None:
    pass


@cli.command(name="create", help="Create a connector object.")
@argument.name
@option.queue(
    help="Queue IDs, that the connector will be associated with. "
    "Required field - will be assigned to an only queue automatically if not specified."
)
@option.service_url
@option.auth_token
@option.params
@option.asynchronous
@click.pass_context
def create_command(
    ctx: click.Context,
    name: str,
    queue_ids: Tuple[int, ...],
    service_url: str,
    auth_token: str,
    params: Optional[str],
    asynchronous: Optional[bool],
) -> None:
    token = auth_token or _generate_token()

    with RossumClient(context=ctx.obj) as rossum:
        if not queue_ids:
            queue_urls = [rossum.get_queue()["url"]]
        else:
            queue_urls = []
            for id_ in queue_ids:
                queue_dict = rossum.get_queue(id_)
                if queue_dict:
                    queue_urls.append(queue_dict["url"])

        response = rossum.create_connector(
            name=name,
            queues=queue_urls,
            service_url=service_url,
            authorization_token=token,
            params=params,
            asynchronous=asynchronous,
        )
        click.echo(f"{response['id']}, {response['name']}, {response['queues']}")


@cli.command(name="list", help="List all connectors.")
@click.pass_context
def list_command(ctx: click.Context,):
    with RossumClient(context=ctx.obj) as rossum:
        connectors_list = rossum.get_connectors((QUEUES,))

    headers = ["id", "name", "service url", "queues", "params", "asynchronous"]

    def get_row(connector: dict) -> List[str]:
        res = [
            connector["id"],
            connector["name"],
            connector["service_url"],
            ", ".join(str(q.get("id", "")) for q in connector["queues"]),
            connector["params"],
            connector["asynchronous"],
        ]
        try:
            token = connector["authorization_token"]
        except KeyError:
            pass
        else:
            res.append(token)
            if "authorization_token" not in headers:
                headers.append("authorization_token")

        return res

    table = [get_row(connector) for connector in connectors_list]

    click.echo(tabulate(table, headers=headers))


@cli.command(name="change", help="Update a connector object.")
@argument.id_
@option.queue(related_object="connector")
@option.name
@option.service_url
@option.auth_token
@option.params
@option.asynchronous
@click.pass_context
def change_command(
    ctx: click.Context,
    id_: str,
    queue_ids: Tuple[int],
    name: Optional[str],
    service_url: str,
    auth_token: str,
    params: Optional[str],
    asynchronous: Optional[bool],
) -> None:
    if not any([queue_ids, service_url, auth_token, params, asynchronous]):
        return

    data: Dict[str, Any] = {}

    with RossumClient(context=ctx.obj) as rossum:
        if queue_ids:
            data["queues"] = [rossum.get_queue(queue)["url"] for queue in queue_ids]
        if name is not None:
            data["name"] = name
        if service_url is not None:
            data["service_url"] = service_url
        if auth_token is not None:
            data["authorization_token"] = auth_token
        if params is not None:
            data["params"] = params
        if asynchronous is not None:
            data["asynchronous"] = asynchronous

        rossum.patch(f"connectors/{id_}", data)


@cli.command(name="delete", help="Delete a connector.")
@argument.id_
@click.confirmation_option(
    prompt="This will delete the connector deployed on the queue. Do you want to continue?"
)
@click.pass_context
def delete_command(ctx: click.Context, id_: str) -> None:
    with RossumClient(context=ctx.obj) as rossum:
        url = rossum.url
        rossum.delete(to_delete={f"{id_}": f"{url}/connectors/{id_}"}, item="connector")


def _generate_token():
    alphabet = string.ascii_letters + string.digits
    return "".join(secrets.choice(alphabet) for _ in range(36))



from typing import Optional, Dict, Any, List, Tuple

import click
from tabulate import tabulate
from rossum import argument, option
from rossum.lib import INBOXES, WORKSPACES, SCHEMAS, USERS, HOOKS
from rossum.lib.api_client import RossumClient, get_json

locale_option = click.option(
    "--locale",
    type=str,
    help="Document locale - passed to the Data Extractor, may influence e.g. date parsing.",
)


@click.group("queue")
def cli() -> None:
    pass


@cli.command(name="create", help="Create queue.")
@click.argument("name")
@option.schema_content(required=True)
@option.email_prefix
@option.bounce_email
@option.workspace_id
@option.connector_id
@locale_option
@option.hook_id
@click.pass_context
def create_command(
    ctx: click.Context,
    name: str,
    schema_content: List[dict],
    email_prefix: Optional[str],
    bounce_email: Optional[str],
    workspace_id: Optional[int],
    connector_id: Optional[int],
    hook_id: Optional[Tuple[int, ...]],
    locale: Optional[str],
) -> None:
    if email_prefix is not None and bounce_email is None:
        raise click.ClickException("Inbox cannot be created without specified bounce email.")

    with RossumClient(context=ctx.obj) as rossum:
        workspace_url = rossum.get_workspace(workspace_id)["url"]
        connector_url = (
            get_json(rossum.get(f"connectors/{connector_id}"))["url"]
            if connector_id is not None
            else None
        )

        hooks_urls = []
        if hook_id:
            for hook in hook_id:
                hook_url = get_json(rossum.get(f"hooks/{hook}"))["url"]
                hooks_urls.append(hook_url)

        schema_dict = rossum.create_schema(f"{name} schema", schema_content)
        queue_dict = rossum.create_queue(
            name, workspace_url, schema_dict["url"], connector_url, hooks_urls, locale
        )

        inbox_dict = {"email": "no email-prefix specified"}
        if email_prefix is not None:
            inbox_dict = rossum.create_inbox(
                f"{name} inbox", email_prefix, bounce_email, queue_dict["url"]
            )
    click.echo(f"{queue_dict['id']}, {inbox_dict['email']}")


@cli.command(name="list", help="List all queues.")
@click.pass_context
def list_command(ctx: click.Context,) -> None:
    with RossumClient(context=ctx.obj) as rossum:
        queues = rossum.get_queues((WORKSPACES, INBOXES, SCHEMAS, USERS, HOOKS))

    table = [
        [
            queue["id"],
            queue["name"],
            str(queue["workspace"].get("id", "")),
            queue["inbox"].get("email", ""),
            str(queue["schema"].get("id", "")),
            ", ".join(str(q.get("id", "")) for q in queue["users"]),
            queue["connector"],
            ", ".join(str(q.get("id", "")) for q in queue["hooks"]),
        ]
        for queue in queues
    ]

    click.echo(
        tabulate(
            table,
            headers=["id", "name", "workspace", "inbox", "schema", "users", "connector", "hooks"],
        )
    )


@cli.command(name="delete", help="Delete a queue.")
@argument.id_
@click.confirmation_option(
    prompt="This will delete ALL DOCUMENTS in the queue. Do you want to continue?"
)
@click.pass_context
def delete_command(ctx: click.Context, id_: int) -> None:
    with RossumClient(context=ctx.obj) as rossum:
        queue = rossum.get_queue(id_)
        rossum.delete({queue["id"]: queue["url"]})


@cli.command(name="change", help="Change a queue.")
@argument.id_
@option.name
@option.schema_content
@option.connector_id
@option.email_prefix
@option.bounce_email
@option.hook_id
@locale_option
@click.pass_context
def change_command(
    ctx: click.Context,
    id_: int,
    name: Optional[str],
    schema_content: Optional[List[dict]],
    email_prefix: Optional[str],
    bounce_email: Optional[str],
    connector_id: Optional[int],
    hook_id: Optional[Tuple[int, ...]],
    locale: Optional[str],
) -> None:

    if not any([name, schema_content, email_prefix, bounce_email, connector_id, locale, hook_id]):
        return

    data: Dict[str, Any] = {}

    if name is not None:
        data["name"] = name

    if locale is not None:
        data["locale"] = locale

    with RossumClient(context=ctx.obj) as rossum:
        if email_prefix or bounce_email:
            queue_dict = rossum.get_queue(id_)
            if not queue_dict["inbox"]:
                inbox_dict = _create_inbox(rossum, queue_dict, email_prefix, bounce_email, name)
            else:
                inbox_dict = _patch_inbox(rossum, queue_dict, email_prefix, bounce_email)
            click.echo(
                f"{inbox_dict['id']}, {inbox_dict['email']}, {inbox_dict['bounce_email_to']}"
            )

        if connector_id is not None:
            data["connector"] = get_json(rossum.get(f"connectors/{connector_id}"))["url"]

        if hook_id:
            hooks_urls = []
            for hook in hook_id:
                hook_url = get_json(rossum.get(f"hooks/{hook}"))["url"]
                hooks_urls.append(hook_url)
                data["hooks"] = hooks_urls

        if schema_content is not None:
            name = name or rossum.get_queue(id_)["name"]
            schema_dict = rossum.create_schema(f"{name} schema", schema_content)
            data["schema"] = schema_dict["url"]

        if data:
            rossum.patch(f"queues/{id_}", data)


def _create_inbox(
    rossum: RossumClient,
    queue_dict: dict,
    email_prefix: Optional[str],
    bounce_email: Optional[str],
    name: Optional[str],
) -> dict:
    return rossum.create_inbox(
        f"{name or queue_dict['name']} inbox", email_prefix, bounce_email, queue_dict["url"]
    )


def _patch_inbox(
    rossum: RossumClient, queue_dict: dict, email_prefix: Optional[str], bounce_email: Optional[str]
) -> dict:
    inbox_data: Dict[str, Any] = {}

    if email_prefix:
        inbox_data["email_prefix"] = email_prefix

    if bounce_email:
        inbox_data["bounce_email_to"] = bounce_email
        inbox_data["bounce_unprocessable_attachments"] = True

    _, inbox_id = queue_dict["inbox"].rsplit("/", 1)
    return get_json(rossum.patch(f"inboxes/{inbox_id}", inbox_data))



#!/usr/bin/env python3
import click as click
from typing import List, Optional, Dict, Union, Callable
from typing.io import IO

from rossum import argument
from rossum.lib.api_client import get_json, RossumClient

SchemaContent = List[dict]
Schema = Dict[str, Union[str, SchemaContent]]
LoadFunction = Callable[[IO[bytes]], SchemaContent]


@click.command(name="update")
@argument.id_(type=str)
@argument.schema_content
@click.option("--rewrite", is_flag=True, type=bool)
@click.option("--name", default=None, type=str)
@click.pass_context
def upload_command(
    ctx: click.Context, id_: str, schema_content: List[dict], rewrite: bool, name: Optional[str]
):
    """
    Update schema in ROSSUM.
    """
    upload_func = _rewrite_schema if rewrite else _create_schema
    with RossumClient(context=ctx.obj) as rossum:
        upload_func(id_, schema_content, rossum, name)


def _rewrite_schema(
    id_: str, schema_content: SchemaContent, rossum: RossumClient, name: Optional[str]
) -> None:
    data: Schema = {"content": schema_content}
    if name is not None:
        data["name"] = name
    rossum.patch(f"schemas/{id_}", data=data)


def _create_schema(
    id_: str, schema_content: SchemaContent, rossum: RossumClient, name: Optional[str]
) -> None:
    original_schema = get_json(rossum.get(f"schemas/{id_}"))
    new_schema = rossum.create_schema(name or original_schema["name"], schema_content)

    for queue_url in original_schema["queues"]:
        if queue_url.startswith(rossum.url):
            queue_url = queue_url[len(rossum.url) + 1 :]
        rossum.patch(queue_url, data={"schema": new_schema["url"]})



import click

from rossum.lib.api_client import RossumClient


@click.group("password")
def cli() -> None:
    pass


@cli.command(name="change", help="Change password of the current user.")
@click.password_option()
@click.pass_context
def change_command(ctx: click.Context, password: str) -> None:
    with RossumClient(context=ctx.obj) as rossum:
        result = rossum.change_user_password(password)
        click.echo(result.get("detail"))
        click.echo('Run "rossum configure" to update existing credentials.')


@cli.command(name="reset", help="Reset password for other user.")
@click.argument("email")
@click.pass_context
def reset_command(ctx: click.Context, email: str) -> None:
    with RossumClient(context=ctx.obj) as rossum:
        result = rossum.reset_user_password(email)
        click.echo(result.get("detail"))



import json
import re
from typing import Optional, Tuple

import click
import jmespath
from typing.io import IO

from rossum.lib.api_client import RossumClient, get_json, get_text
from rossum import option


@click.command(name="extract", help="Upload documents and extract data from them.")
@click.argument("queue_id", metavar="QUEUE_ID", type=str)
@click.argument(
    "files", metavar="FILE_TO_UPLOAD", nargs=-1, type=click.Path(readable=True), required=True
)
@click.option(
    "--format",
    "format_",
    default="json",
    type=click.Choice(["json", "csv", "xml"]),
    help="Format of the output file.",
)
@option.output_file
@click.option("--indent", default=4, type=int, help="Set indentation for JSON output file.")
@click.option(
    "--ensure-ascii",
    is_flag=True,
    type=bool,
    help="Ensure that the output file is in valid ASCII characters.",
)
@click.pass_context
def get_data(
    ctx: click.Context,
    queue_id: int,
    files: Tuple[str],
    indent: int,
    ensure_ascii: bool,
    format_: str,
    output_file: Optional[IO[str]],
):
    annotations_to_export = list()

    with RossumClient(context=ctx.obj) as rossum:
        for file in files:
            json_response = rossum.upload_document(queue_id, file)
            annotation_id = get_id(json_response)
            annotations_to_export.append(annotation_id)
            rossum.poll_annotation(annotation_id, _is_done)

        export_data = rossum.export_data(queue_id, annotations_to_export, format_)

        if format_ == "json":
            output = json.dumps(get_json(export_data), indent=indent, ensure_ascii=ensure_ascii)
        else:
            output = get_text(export_data)
        click.echo(output.encode("utf-8"), file=output_file, nl=False)


def get_id(json_response: dict) -> int:
    annotation_url = jmespath.search("results[*].annotation | [0]", json_response)
    annotation_id = re.search(r"(?<=annotations/)\d*$", annotation_url)
    if annotation_id is not None:
        annotation_id_number = int(annotation_id.group())
        return annotation_id_number
    raise click.ClickException("Could not get annotation ID from the response.")


def _is_done(response):
    return response["status"] != "importing"



#!/usr/bin/env python3
import difflib
import json
import pprint

import click
import re
from jsondiff import diff as json_diff, JsonDumper
from typing import Iterable, IO

from rossum.lib import split_dict_params

DIFF_TYPES = ["jsondiff", "difflib"]
DIFF_RE = re.compile(r"^[+-^]")


@click.command("compare", help="Compare 2 json files.")
@click.argument("json1", type=click.File("rb"))
@click.argument("json2", type=click.File("rb"))
@click.option("--method", "-m", type=click.Choice(DIFF_TYPES), default=DIFF_TYPES[0])
@click.option("--option", "-o", "options", multiple=True, type=str)
def cli(json1: IO[str], json2: IO[str], method: str, options: Iterable[str]) -> None:
    try:
        options_dict = dict(split_dict_params(options))
    except ValueError as e:
        raise click.BadArgumentUsage("Expecting <key>=<value> pairs.") from e

    d1 = json.load(json1)
    d2 = json.load(json2)
    if method == "difflib":
        diff = _difflib(d1, d2, **options_dict)
    elif method == "jsondiff":
        diff = _json_diff(d1, d2, **options_dict)
    else:
        raise NotImplementedError

    click.echo(diff)


def _difflib(d1: dict, d2: dict, **kwargs) -> str:
    diff_list = difflib.ndiff(pprint.pformat(d1).splitlines(), pprint.pformat(d2).splitlines())
    if str(kwargs.get("fulldiff", False)).lower() in ("true", "t", "1"):
        return "\n" + "\n".join(diff_list)
    else:
        return "\n" + "\n".join(d for d in diff_list if DIFF_RE.match(d))


def _json_diff(d1: dict, d2: dict, **kwargs) -> str:
    dumper = JsonDumper(indent=2)
    kwargs = {"syntax": "symmetric", **kwargs}
    return json_diff(d1, d2, **kwargs, dumper=dumper, dump=True)



from typing import Optional, Dict, Any

import click
from tabulate import tabulate

from rossum import argument, option
from rossum.lib import QUEUES
from rossum.lib.api_client import RossumClient


@click.group("workspace")
def cli() -> None:
    pass


@cli.command(name="create", short_help="Create workspace.")
@click.argument("name")
@option.organization
@click.pass_context
def create_command(ctx: click.Context, name: str, organization_id: Optional[int]) -> None:
    with RossumClient(context=ctx.obj) as rossum:
        organization_url = rossum.get_organization(organization_id)["url"]

        workspace_response = rossum.create_workspace(name, organization_url)

    click.echo(workspace_response["id"])


@cli.command(name="list", help="List all workspaces.")
@click.pass_context
def list_command(ctx: click.Context,):
    with RossumClient(context=ctx.obj) as rossum:
        workspaces = rossum.get_workspaces((QUEUES,))

    table = [
        [
            workspace["id"],
            workspace["name"],
            ", ".join(str(q.get("id", "")) for q in workspace["queues"]),
        ]
        for workspace in workspaces
    ]

    click.echo(tabulate(table, headers=["id", "name", "queues"]))


@cli.command(name="delete", help="Delete a workspace.")
@argument.id_
@click.confirmation_option()
@click.pass_context
def delete_command(ctx: click.Context, id_: int) -> None:
    with RossumClient(context=ctx.obj) as rossum:
        workspace = rossum.get_workspace(id_)
        queues = rossum.get_queues(workspace=workspace["id"])
        documents = {}
        for queue in queues:
            res, _ = rossum.get_paginated(
                "annotations",
                {"page_size": 50, "queue": queue["id"], "sideload": "documents"},
                key="documents",
            )
            documents.update({d["id"]: d["url"] for d in res})

        rossum.delete({workspace["id"]: workspace["url"], **documents})


@cli.command(name="change", help="Change a workspace.")
@argument.id_
@option.name
@click.pass_context
def change_command(ctx: click.Context, id_: str, name: Optional[str]) -> None:
    if not any([name]):
        return

    data: Dict[str, Any] = {}
    if name is not None:
        data["name"] = name

    with RossumClient(context=ctx.obj) as rossum:
        rossum.patch(f"workspaces/{id_}", data)



from typing import Tuple, Optional, Dict, Any, List

import click
from pathlib import Path
from rossum import option, argument
from rossum.lib import QUEUES
from rossum.lib.api_client import RossumClient
from tabulate import tabulate


def cleanup_config(config: Dict) -> Dict:
    new_config = {}
    for key, value in config.items():
        if key.startswith("config_"):
            new_key = key.replace("config_", "")
            new_config[new_key] = value
    config = {key: value for key, value in new_config.items() if value is not None}
    if "code" in config.keys():
        config["code"] = Path(config["code"]).read_text()
    return config


@click.group("hook")
def cli() -> None:
    pass


@cli.command(name="create", help="Create a hook object.")
@argument.name
@option.hook_type
@option.queue(
    help="Queue IDs, that the hook will be associated with. "
    "Required field - will be assigned to an only queue automatically if not specified."
)
@option.active
@option.events
@option.config_url
@option.config_insecure_ssl
@option.config_secret
@option.config_code
@option.config_runtime
@option.sideload
@click.pass_context
def create_command(
    ctx: click.Context,
    name: str,
    hook_type: str,
    queue_ids: Tuple[int, ...],
    active: bool,
    events: Tuple[str, ...],
    sideload: Tuple[str, ...],
    **kwargs,
) -> None:

    with RossumClient(context=ctx.obj) as rossum:
        if not queue_ids:
            queue_urls = [rossum.get_queue()["url"]]
        else:
            queue_urls = []
            for id_ in queue_ids:
                queue_dict = rossum.get_queue(id_)
                if queue_dict:
                    queue_urls.append(queue_dict["url"])

        config = {**kwargs}
        config = cleanup_config(config)

        response = rossum.create_hook(
            name=name,
            hook_type=hook_type,
            queues=queue_urls,
            active=active,
            events=list(events),
            sideload=list(sideload),
            config=config,
        )

        additional_fields = [
            value
            for key, value in response["config"].items()
            if key not in ["code", "runtime", "insecure_ssl"]
        ]

        regular_fields = f"{response['id']}, {response['name']}, {response['queues']}, {response['events']}, {response['sideload']}"
        click.echo(
            regular_fields + ", " + f"{', '.join(map(str, additional_fields))}"
            if additional_fields != []
            else regular_fields
        )


@cli.command(name="list", help="List all hooks.")
@click.pass_context
def list_command(ctx: click.Context,):
    with RossumClient(context=ctx.obj) as rossum:
        hooks_list = rossum.get_hooks((QUEUES,))

    headers = ["id", "name", "events", "queues", "active", "sideload"]

    def get_row(hook: dict) -> List[str]:
        fields = [
            hook["id"],
            hook["name"],
            ", ".join(e for e in hook["events"]),
            ", ".join(str(q.get("id", "")) for q in hook["queues"]),
            hook["active"],
            hook["sideload"],
        ]

        additional = ["url", "insecure_ssl", "secret"]

        for field in additional:
            if field in hook["config"]:
                fields.append(hook["config"][field])

        for header in additional:
            if header not in headers:
                headers.append(header)

        hook_list = [item for item in fields]
        return hook_list

    table = [get_row(hook) for hook in hooks_list]

    click.echo(tabulate(table, headers=headers))


@cli.command(name="change", help="Update a hook object.")
@argument.id_
@option.queue(related_object="hook")
@option.name
@option.hook_type
@option.events
@option.active
@option.config_url
@option.config_secret
@option.config_insecure_ssl
@option.config_code
@option.config_runtime
@option.sideload
@click.pass_context
def change_command(
    ctx: click.Context,
    id_: int,
    queue_ids: Tuple[int, ...],
    name: Optional[str],
    hook_type: str,
    events: Tuple[str, ...],
    active: Optional[bool],
    sideload: Tuple[str, ...],
    **kwargs,
) -> None:

    config = {**kwargs}
    config = cleanup_config(config)

    if not any([queue_ids, name, active, events, sideload, config]):
        return

    data: Dict[str, Any] = {"config": {}}

    with RossumClient(context=ctx.obj) as rossum:
        if queue_ids:
            data["queues"] = [rossum.get_queue(queue)["url"] for queue in queue_ids]
        if name is not None:
            data["name"] = name
        if hook_type:
            data["type"] = hook_type
        if active is not None:
            data["active"] = active
        if events:
            data["events"] = list(events)
        if config:
            data["config"] = config
        if sideload:
            data["sideload"] = list(sideload)

        rossum.patch(f"hooks/{id_}", data)


@cli.command(name="delete", help="Delete a hook.")
@argument.id_
@click.confirmation_option(
    prompt="This will delete the hook deployed on the queue. Do you want to continue?"
)
@click.pass_context
def delete_command(ctx: click.Context, id_: str) -> None:
    with RossumClient(context=ctx.obj) as rossum:
        url = rossum.url
        rossum.delete(to_delete={f"{id_}": f"{url}/hooks/{id_}"}, item="hook")



import json
import sys
from contextlib import AbstractContextManager
from pathlib import PurePath
from platform import platform

import click
import polling2
import requests
from requests import Response
from typing import Any, BinaryIO, Callable, Dict, IO, Iterable, List, Optional, Tuple, Union

from rossum import __version__, CTX_PROFILE, CTX_DEFAULT_PROFILE
from rossum.configure import get_credential
from . import (
    ORGANIZATIONS,
    APIObject,
    WORKSPACES,
    QUEUES,
    SCHEMAS,
    CONNECTORS,
    HOOKS,
    USERS,
    GROUPS,
    ANNOTATIONS,
)


class RossumException(click.ClickException):
    pass


RequestsFiles = Dict[str, Tuple[Optional[str], Union[IO[bytes], BinaryIO, str]]]

HEADERS = {"User-Agent": f"rossum/{__version__} ({platform()})"}


class APIClient(AbstractContextManager):
    def __init__(
        self,
        context: Optional[dict],
        url: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        use_api_version: bool = True,
        auth_using_token: bool = True,
        max_token_lifetime: Optional[int] = None,
    ):
        self._url = url
        self._user = user
        self._password = password
        self._use_api_version = use_api_version
        self._auth_using_token = auth_using_token
        self._max_token_lifetime = max_token_lifetime
        self._profile = (context or {}).get(CTX_PROFILE, CTX_DEFAULT_PROFILE)

        self.token: Optional[str] = None
        self.timeout: Optional[float] = None

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.logout()

    @classmethod
    def csv(
        cls, context: Optional[dict], url: str = None, user: str = None, password: str = None
    ) -> "APIClient":
        return cls(context, url, user, password, False, False)

    @property
    def user(self) -> str:
        if self._user is None:
            self._user = get_credential("username", self._profile)
        return self._user

    @property
    def password(self) -> str:
        if self._password is None:
            self._password = get_credential("password", self._profile)
        return self._password

    @property
    def url(self) -> str:
        if self._url is None:
            _url = get_credential("url", self._profile).rstrip("/")
            self._url = f'{_url}{"/v1" if self._use_api_version else ""}'
        return self._url

    def get_token(self) -> str:
        # self.post cannot be used as it is dependent on self.get_token().
        login_data: Dict[str, Union[str, int]] = {"username": self.user, "password": self.password}
        if self._max_token_lifetime:
            login_data["max_token_lifetime_s"] = self._max_token_lifetime
        response = requests.post(f"{self.url}/auth/login", json=login_data, headers=HEADERS)
        if response.status_code == 401:
            raise RossumException(f"Login failed with the provided credentials.")
        elif not response.ok:
            raise RossumException(f"Invalid response [{response.url}]: {response.text}")

        return response.json()["key"]

    def post(
        self,
        path: Union[str, APIObject],
        data: dict = None,
        expected_status_code: int = 201,
        files: Optional[RequestsFiles] = None,
    ) -> Response:
        return self._request_url(
            "post",
            f"{self.url}/{path}",
            json=data,
            expected_status_code=expected_status_code,
            files=files,
        )

    def patch(self, path: Union[str, APIObject], data: dict) -> Response:
        return self._request_url("patch", f"{self.url}/{path}", json=data)

    def get(self, path: Union[str, APIObject], query: dict = None) -> Response:
        return self._request_url("get", f"{self.url}/{path}", query)

    def get_url(self, url: str, query: dict = None) -> Response:
        return self._request_url("get", url, query)

    def delete_url(self, url: str) -> Response:
        return self._request_url("delete", url, expected_status_code=204)

    def _do_request(self, method: str, url: str, query: dict = None, **kwargs) -> Response:
        auth = self._authentication
        headers = {**HEADERS, **auth.pop("headers", {}), **kwargs.pop("headers", {})}
        response = requests.request(
            method,
            url,
            params=_encode_booleans(query),
            headers=headers,
            timeout=self.timeout,
            **auth,
            **kwargs,
        )
        return response

    def _request_url(
        self, method: str, url: str, query: dict = None, expected_status_code: int = 200, **kwargs
    ) -> Response:
        response = self._do_request(method, url, query, **kwargs)
        if response.status_code != expected_status_code:
            raise RossumException(f"Invalid response [{response.url}]: {response.text}")
        return response

    def delete(self, to_delete: Dict[str, str], verbose: int = 0, item: str = "annotation") -> None:
        for id_, url in to_delete.items():
            try:
                self.delete_url(url)
            except RossumException as exc:
                click.echo(f'Deleting {item} {id_} caused "{exc}".')
            except Exception as exc:
                click.echo(f'Deleting {item} {id_} caused an unexpected exception: "{exc}".')
                raise RossumException(str(exc))
            else:
                if verbose > 1:
                    click.echo(f"Deleted {item} {id_}.")

    def get_paginated(
        self,
        path: Union[str, APIObject],
        query: Optional[Dict[str, Any]] = None,
        *,
        key: str = "results",
    ) -> Tuple[List[Dict[str, Any]], int]:
        response = self.get(path, query)
        response_dict = response.json()

        res = response_dict[key]
        next_page = response_dict["pagination"]["next"]

        while next_page:
            response = self.get_url(next_page)
            response_dict = response.json()

            res.extend(response_dict[key])
            next_page = response_dict["pagination"]["next"]

        return res, response_dict["pagination"]["total"]

    def _sideload(
        self, objects: List[dict], sideloads: Optional[Iterable[APIObject]] = None
    ) -> List[dict]:
        for sideload in sideloads or []:
            sideloaded, _ = self.get_paginated(sideload)
            sideloaded_dicts = {
                sideloaded_dict["url"]: sideloaded_dict for sideloaded_dict in sideloaded
            }

            def inject_sideloaded(obj: dict) -> dict:
                try:
                    url = obj[sideload.singular]
                except KeyError:
                    obj[sideload.plural] = [
                        sideloaded_dicts[url]
                        for url in obj[sideload.plural]
                        if url in sideloaded_dicts
                    ]
                else:
                    obj[sideload.singular] = sideloaded_dicts.get(url, {})
                return obj

            objects = [inject_sideloaded(o) for o in objects]
        return objects

    @property
    def _authentication(self) -> dict:
        if self._auth_using_token:
            if self.token is None:
                self.token = self.get_token()
            return {"headers": {"Authorization": "Token " + self.token}}
        else:
            return {"auth": (self.user, self.password)}

    def logout(self) -> None:
        if self._auth_using_token:
            self.post("auth/logout", {}, expected_status_code=200)


class RossumClient(APIClient):
    def get_organization(self, organization_id: Optional[int] = None) -> dict:
        if organization_id is None:
            user_details = self.get_user()
            try:
                organization_url = user_details[ORGANIZATIONS.singular]
            except KeyError:
                organization_url = get_json(self.get_url(user_details["url"]))[
                    ORGANIZATIONS.singular
                ]
            res = self.get_url(organization_url)
        else:
            res = self.get(f"{ORGANIZATIONS}/{organization_id}")
        return get_json(res)

    def get_workspaces(
        self, sideloads: Optional[Iterable[APIObject]] = None, *, organization: Optional[int] = None
    ) -> List[dict]:
        query = {}
        if organization:
            query[ORGANIZATIONS.singular] = organization
        workspaces_list, _ = self.get_paginated(WORKSPACES, query=query)
        self._sideload(workspaces_list, sideloads)
        return workspaces_list

    def get_workspace(
        self, id_: Optional[int] = None, sideloads: Optional[Iterable[APIObject]] = None
    ) -> dict:
        if id_ is None:
            try:
                [workspace] = self.get_workspaces()
            except ValueError as e:
                raise RossumException("Workspace ID must be specified.") from e
        else:
            workspace = get_json(self.get(f"{WORKSPACES}/{id_}"))

        self._sideload([workspace], sideloads)
        return workspace

    def get_queues(
        self,
        sideloads: Optional[Iterable[APIObject]] = None,
        *,
        workspace: Optional[int] = None,
        users: Optional[Iterable[int]] = None,
        hooks: Optional[Iterable[int]] = None,
    ) -> List[dict]:
        query: Dict[str, Any] = {}
        if workspace:
            query[WORKSPACES.singular] = workspace
        if users:
            query[USERS.plural] = users
        if hooks:
            query[HOOKS.plural] = hooks
        queues_list, _ = self.get_paginated(QUEUES, query=query)
        self._sideload(queues_list, sideloads)
        return queues_list

    def get_queue(
        self, id_: Optional[int] = None, sideloads: Optional[Iterable[APIObject]] = None
    ) -> dict:
        if id_ is None:
            try:
                [queue] = self.get_queues()
            except ValueError as e:
                raise RossumException("Queue ID must be specified.") from e
        else:
            queue = get_json(self.get(f"{QUEUES}/{id_}"))

        self._sideload([queue], sideloads)
        return queue

    def get_users(
        self,
        sideloads: Optional[Iterable[APIObject]] = None,
        *,
        username: Optional[str] = None,
        is_active: Optional[bool] = None,
    ) -> List[dict]:
        query: Dict[str, Union[str, bool]] = {}
        if username:
            query["username"] = username
        if is_active is not None:
            query["is_active"] = is_active
        users_list, _ = self.get_paginated(USERS, query=query)
        self._sideload(users_list, sideloads)
        return users_list

    def get_user(self, id_: Optional[int] = None) -> dict:
        if id_ is None:
            user = get_json(self.get("auth/user"))
        else:
            user = get_json(self.get(f"{USERS}/{id_}"))
        return user

    def get_groups(self, *, group_name: Optional[str]) -> List[dict]:
        if group_name is None:
            return []
        groups_list, _ = self.get_paginated(GROUPS, query={"name": group_name})
        return groups_list

    def get_connectors(self, sideloads: Optional[Iterable[APIObject]] = None) -> List[dict]:
        connectors_list, _ = self.get_paginated(CONNECTORS)
        self._sideload(connectors_list, sideloads)
        return connectors_list

    def get_hooks(
        self, sideloads: Optional[Iterable[APIObject]] = None, query: Dict = None
    ) -> List[dict]:
        hooks_list, _ = self.get_paginated(HOOKS, query)
        self._sideload(hooks_list, sideloads)
        return hooks_list

    def get_annotation(self, id_: Optional[int] = None) -> dict:
        if id_ is None:
            raise RossumException("Annotation ID wasn't specified.")
        return get_json(self.get(f"{ANNOTATIONS}/{id_}"))

    def poll_annotation(
        self, annotation: int, check_success: Callable, max_retries=120, sleep_secs=5
    ) -> dict:
        return polling2.poll(
            lambda: self._get_annotation_polling(annotation),
            check_success=check_success,
            step=sleep_secs,
            timeout=int(round(max_retries * sleep_secs)),
        )

    def _get_annotation_polling(self, annotation: int) -> dict:
        annotation_object = self.get_annotation(annotation)
        status = annotation_object["status"]
        annotation_path = annotation_object["url"]
        if status == "importing":
            click.echo(".", nl=False, err=True)
            sys.stdout.flush()
        elif status == "to_review":
            click.echo(f"Processing of the annotation at {annotation_path} finished.", err=True)
        elif status == "failed_import":
            click.echo(" Processing failed.")
        return annotation_object

    def create_workspace(
        self, name: str, organization: str, metadata: Optional[Dict] = None
    ) -> Dict[str, Any]:
        data = {"name": name, "organization": organization}  # type: Dict[str, Any]
        if metadata:
            data.update({"metadata": metadata})
        return get_json(self.post("workspaces", data))

    def create_schema(self, name: str, content: List[dict]) -> dict:
        return get_json(self.post(SCHEMAS, data={"name": name, "content": content}))

    def create_queue(
        self,
        name: str,
        workspace_url: str,
        schema_url: str,
        connector_url: Optional[str] = None,
        hooks_urls: Optional[List] = None,
        locale: Optional[str] = None,
        rir_url: str = "https://all.rir.rossum.ai",
    ) -> dict:
        data = {"name": name, "workspace": workspace_url, "schema": schema_url, "rir_url": rir_url}
        if connector_url is not None:
            data[CONNECTORS.singular] = connector_url
        if hooks_urls is not None:
            data[HOOKS.plural] = hooks_urls  # type: ignore
        if locale is not None:
            data["locale"] = locale
        return get_json(self.post("queues", data))

    def create_inbox(
        self, name: str, email_prefix: Optional[str], bounce_email: Optional[str], queue_url: str
    ) -> dict:
        if not (email_prefix and bounce_email):
            raise RossumException(
                "Inbox cannot be created without both bounce email and email prefix specified."
            )

        return get_json(
            self.post(
                "inboxes",
                data={
                    "name": name,
                    "email_prefix": email_prefix,
                    "bounce_email_to": bounce_email,
                    "bounce_unprocessable_attachments": True,
                    "queues": [queue_url],
                },
            )
        )

    def create_user(
        self,
        username: str,
        organization: str,
        queues: List[str],
        password: str,
        group: str,
        locale: str,
    ) -> dict:
        return get_json(
            self.post(
                USERS,
                data={
                    "username": username,
                    "email": username,
                    "organization": organization,
                    "password": password,
                    "groups": [g["url"] for g in self.get_groups(group_name=group)],
                    "queues": queues,
                    "ui_settings": {"locale": locale},
                },
            )
        )

    def change_user_password(self, new_password: str) -> dict:
        data = {
            "new_password1": new_password,
            "new_password2": new_password,
            "old_password": self.password,
        }
        return get_json(self.post("auth/password/change", data, expected_status_code=200))

    def reset_user_password(self, email: str) -> dict:
        data = {"email": email}
        return get_json(self.post("auth/password/reset", data, expected_status_code=200))

    def create_connector(
        self,
        name: str,
        queues: List[str],
        service_url: str,
        authorization_token: str = None,
        params: Optional[str] = None,
        asynchronous: Optional[bool] = True,
    ) -> dict:
        data = {
            "name": name,
            "queues": queues,
            "service_url": service_url,
            "authorization_token": authorization_token,
            "params": params,
            "asynchronous": asynchronous,
        }
        return get_json(self.post("connectors", data))

    def create_hook(
        self,
        name: str,
        hook_type: str,
        queues: List[str],
        active: bool,
        events: List[str],
        sideload: List[str],
        config: Dict,
    ) -> dict:

        data = {
            "name": name,
            "type": hook_type,
            "queues": queues,
            "active": active,
            "events": events,
            "sideload": sideload,
            "config": config,
        }
        return get_json(self.post("hooks", data))

    def upload_document(
        self, id_: int, file: str, filename_overwrite: str = "", values: Dict[str, str] = None
    ) -> dict:
        filename = PurePath(filename_overwrite).name or PurePath(file).name
        files: RequestsFiles = {"content": (filename, open(f"{file}", "rb"))}
        if values is not None:
            files["values"] = (None, json.dumps(values))
        return get_json(self.post(f"queues/{id_}/upload", files=files))

    def set_metadata(self, object_type: APIObject, object_id: int, metadata: Dict[str, Any]):
        return get_json(self.patch(f"{object_type}/{object_id}", {"metadata": metadata}))

    def export_data(self, id_: int, annotation_ids: Iterable[int], format_: str) -> Response:
        ids = ",".join(str(a) for a in annotation_ids)
        return self.get(f"queues/{id_}/export", query={"id": ids, "format": format_})

    def get_schemas(self, sideloads: Optional[Iterable[APIObject]] = None) -> List[dict]:
        schemas_list, _ = self.get_paginated(SCHEMAS)
        self._sideload(schemas_list, sideloads)
        return schemas_list


def get_json(response: Response) -> dict:
    try:
        return response.json()
    except ValueError as e:
        raise RossumException(f"Invalid JSON [{response.url}]: {response.text}") from e


def get_text(response: Response) -> str:
    try:
        return response.text
    except ValueError as e:
        raise RossumException(f"Invalid text [{response.url}]: {response.text}") from e


def _encode_booleans(query: Optional[dict]) -> Optional[dict]:
    if query is None:
        return query

    def bool_to_str(b: Any) -> Any:
        if isinstance(b, bool):
            return str(b).lower()
        return b

    res = {}
    for k, vs in query.items():
        if isinstance(vs, str) or not hasattr(vs, "__iter__"):
            vs = [vs]
        res[k] = (bool_to_str(v) for v in vs)
    return res



import json
import secrets
import string
from contextlib import suppress
from dataclasses import dataclass

from typing import Iterable, Iterator, Tuple, Union, Dict, List

# todo: use TypedDict if available https://www.python.org/dev/peps/pep-0589/
# todo: use Recursion (see the following) when https://github.com/python/mypy/issues/731 is ready
# DataPointDictItem = Union[str, int, "DataPointDict", None, "DataPoints"]
DataPointDictItem = Union[str, int, dict, None, list]
DataPointDict = Dict[str, DataPointDictItem]
DataPoints = List[DataPointDict]


def split_dict_params(
    datapoint_parameters: Iterable[str]
) -> Iterator[Tuple[str, DataPointDictItem]]:
    for param in datapoint_parameters:
        key, value = param.split("=", 1)
        with suppress(ValueError):
            value = json.loads(value)
        yield key, value


def generate_secret(length: int = 10):
    alphabet = string.ascii_letters + string.digits
    return "".join(secrets.choice(alphabet) for _ in range(length))


@dataclass(frozen=True)
class APIObject:
    NOT_SET = "NOT_SET"

    plural: str
    singular: str = NOT_SET

    def __post_init__(self) -> None:
        if self.singular is self.NOT_SET:
            # https://docs.python.org/3/library/dataclasses.html#frozen-instances
            object.__setattr__(self, "singular", self.plural.rstrip("s"))

    def __str__(self) -> str:
        return self.plural

    # todo: add list and detail methods (the base_url needs to be somehow obtained)


ORGANIZATIONS = APIObject("organizations")
WORKSPACES = APIObject("workspaces")
QUEUES = APIObject("queues")
INBOXES = APIObject("inboxes", "inbox")
CONNECTORS = APIObject("connectors")
HOOKS = APIObject("hooks")
SCHEMAS = APIObject("schemas")
USERS = APIObject("users")
GROUPS = APIObject("groups")
ANNOTATIONS = APIObject("annotations")
PAGES = APIObject("pages")



from itertools import chain

from tabulate import tabulate
from typing import Tuple, Optional, List, Dict

import click
from rossum import option
from rossum.lib import USERS, QUEUES
from rossum.lib.api_client import RossumClient


@click.group("user_assignment")
def cli() -> None:
    """Assignment of users to queues"""
    pass  # pragma: no cover


@cli.command(name="list")
@option.user(required=False, help="User IDs, which the queues will be filtered by.")
@option.queue(required=False, help="Queue IDs, which the users will be filtered by.")
@click.pass_context
def list_command(ctx: click.Context, user_ids: Tuple[int], queue_ids: Tuple[int]) -> None:
    """List all users and their assignments to queues."""
    with RossumClient(context=ctx.obj) as rossum:
        queue_users = rossum.get_queues((USERS,), users=user_ids)

    user_queues: Dict[int, List[List[Optional[str]]]] = {}
    for queue in queue_users:
        if queue_ids and int(queue["id"]) not in queue_ids:
            continue
        for user in queue["users"]:
            user_id = int(user["id"])
            if user_ids and user_id not in user_ids:
                continue

            if user_id not in user_queues:
                user_queues[user_id] = [[user["id"], user["username"], queue["id"], queue["name"]]]
            else:
                user_queues[user_id].append([None, None, queue["id"], queue["name"]])
    user_queues = dict(sorted(user_queues.items()))
    click.echo(
        tabulate(
            chain.from_iterable(user_queues.values()),
            headers=["id", "username", "queue id", "queue name"],
        )
    )


@cli.command(name="add", help="Add user to queues.")
@option.user
@option.queue
@click.pass_context
def add_command(ctx: click.Context, user_ids: Tuple[int], queue_ids: Tuple[int]) -> None:
    with RossumClient(context=ctx.obj) as rossum:
        for user_id in user_ids:
            user = rossum.get_user(user_id)
            new_queues = user["queues"] + [rossum.get_queue(q_id)["url"] for q_id in queue_ids]
            rossum.patch(f"{USERS}/{user_id}", {str(QUEUES): new_queues})


@cli.command(name="remove", help="Remove user from queues.")
@option.user
@option.queue
@click.pass_context
def remove_command(ctx: click.Context, user_ids: Tuple[int], queue_ids: Tuple[int]) -> None:
    with RossumClient(context=ctx.obj) as rossum:
        for user_id in user_ids:
            queues = rossum.get_queues(users=[user_id])
            new_queues = [q["url"] for q in queues if int(q["id"]) not in queue_ids]
            rossum.patch(f"{USERS}/{user_id}", {str(QUEUES): new_queues})



import json

from typing import Optional, IO, Tuple

import click as click

from rossum import option, argument
from rossum.lib import DataPoints, DataPointDict
from . import (
    traverse_datapoints,
    substitute_options,
    remove,
    wrap_in_multivalue,
    add,
    _new_datapoint,
    change,
    get,
)


@click.group("transform", help="Transform schema file content.")
@click.pass_context
@click.option(
    "--indent", default=2, type=int, show_default=True, help="Indentation of resulting JSON."
)
@click.option(
    "--ensure-ascii", is_flag=True, type=bool, help="Escape non-ASCII characters in resulting JSON."
)
@click.option("--sort-keys", is_flag=True, type=bool, help="Order keys in resulting JSON.")
@option.output_file
def cli(
    ctx: click.Context,
    indent: int,
    ensure_ascii: bool,
    sort_keys: bool,
    output_file: Optional[IO[str]],
) -> None:
    pass


@cli.command(name="substitute-options", help="Substitute options in existing enum datapoint.")
@click.pass_context
@argument.schema_file
@argument.id_(type=str)
@click.argument("new_options", type=click.File("rb"))
def substitute_options_command(ctx: click.Context, new_options: IO[str], id_: str) -> DataPoints:
    options_dict = json.load(new_options)
    return traverse_datapoints(ctx.obj["SCHEMA"], substitute_options, id_=id_, options=options_dict)


@cli.command(name="remove", help="Remove datapoints.")
@click.pass_context
@argument.schema_file
@click.argument("ids", nargs=-1, type=str)
def remove_command(ctx: click.Context, ids: Tuple[str, ...]) -> DataPoints:
    return traverse_datapoints(ctx.obj["SCHEMA"], remove, ids=ids)


@cli.command(
    name="wrap-in-multivalue",
    short_help="Put all datapoints into a multivalue.",
    help="Put all datapoints into a multivalue (unless they are already in a multivalue).",
)
@click.pass_context
@argument.schema_file
@click.argument("exclude_ids", nargs=-1, type=str)
def wrap_in_multivalue_command(ctx: click.Context, exclude_ids: Tuple[str, ...]) -> DataPoints:
    return traverse_datapoints(ctx.obj["SCHEMA"], wrap_in_multivalue, exclude_ids=set(exclude_ids))


@cli.command(
    name="add",
    short_help="Create new datapoint.",
    help="""
Create new datapoint.

DATAPOINT_PARAMETERS are expected as <key>=<value> pairs, where <value> can be a json.
""",
)
@click.pass_context
@argument.schema_file
@click.argument("parent_id", type=str)
@argument.datapoint_parameters
@click.option(
    "--place-before",
    "-p",
    type=str,
    default=None,
    help="Id of datapoint which will follow the added datapoint.",
)
def add_command(
    ctx: click.Context,
    parent_id: str,
    datapoint_parameters: DataPointDict,
    place_before: Optional[str],
) -> DataPoints:

    return traverse_datapoints(
        ctx.obj["SCHEMA"],
        add,
        parent_id=parent_id,
        datapoint_to_add=_new_datapoint(datapoint_parameters),
        place_before=place_before,
    )


@cli.command(
    name="change",
    short_help="Change existing datapoint.",
    help="""
Change existing datapoint.

ID can be set to ALL, then all datapoints are changed.
DATAPOINT_PARAMETERS are expected as <key>=<value> pairs, where <value> can be a JSON.
""",
)
@click.pass_context
@argument.schema_file
@argument.id_(type=str)
@argument.datapoint_parameters
@click.option(
    "-c",
    "--category",
    "categories",
    type=click.Choice(["datapoint", "multivalue", "tuple", "section"]),
    multiple=True,
    help="Change only datapoints of specified categories. Useful with <id> set to ALL. "
    "Multiple categories can be set.",
)
def change_command(
    ctx: click.Context, id_: str, datapoint_parameters: DataPointDict, categories: Tuple[str]
) -> DataPoints:
    return traverse_datapoints(
        ctx.obj["SCHEMA"],
        change,
        id_=id_,
        to_change=datapoint_parameters,
        filtered_categories=categories,
    )


@cli.command(name="move", help="Move datapoint to new parent datapoint.")
@click.pass_context
@argument.schema_file
@click.argument("source_id", type=str)
@click.argument("target_id", type=str)
def move_command(ctx: click.Context, source_id: str, target_id: str) -> DataPoints:
    source_dp = get(ctx.obj["SCHEMA"], id_=source_id)
    new_schema = traverse_datapoints(ctx.obj["SCHEMA"], remove, ids=(source_id,))
    return traverse_datapoints(new_schema, add, parent_id=target_id, datapoint_to_add=source_dp)


@cli.resultcallback()
@click.pass_context
def process_result(
    ctx: click.Context,
    result: DataPoints,
    indent: int,
    ensure_ascii: bool,
    sort_keys: bool,
    output_file: Optional[IO[str]],
) -> None:
    output = json.dumps(result, indent=indent, ensure_ascii=ensure_ascii, sort_keys=sort_keys)
    click.echo(output.encode("utf-8"), file=output_file, nl=False)



import json

from typing import Optional, IO

import click
from tabulate import tabulate

from rossum import option, argument
from rossum.lib.api_client import APIClient, get_json, RossumClient
from rossum.lib import QUEUES

from . import upload
from .xlsx import SchemaToXlsx
from .transform import commands as transform


@click.group("schema")
def cli() -> None:
    pass


cli.add_command(transform.cli)
cli.add_command(upload.upload_command)


@cli.command(name="get", help="Download schema from ROSSUM.")
@click.pass_context
@argument.id_(type=str)
@click.option("--indent", default=2, type=int)
@click.option("--ensure-ascii", is_flag=True, type=bool)
@click.option("--format", "format_", default="json", type=click.Choice(["json", "xlsx"]))
@option.output_file
def download_command(
    ctx: click.Context,
    id_: str,
    indent: int,
    ensure_ascii: bool,
    format_: str,
    output_file: Optional[IO[str]],
):
    with APIClient(context=ctx.obj) as api_client:
        schema_dict = get_json(api_client.get(f"schemas/{id_}"))
    if format_ == "json":
        schema_file = json.dumps(
            schema_dict["content"], indent=indent, ensure_ascii=ensure_ascii, sort_keys=True
        ).encode("utf-8")
    else:
        schema_file = SchemaToXlsx().convert(schema_dict["content"])

    click.echo(schema_file, file=output_file, nl=False)


@cli.command(name="list", help="List all schemas.")
@click.pass_context
def list_command(ctx: click.Context,):
    with RossumClient(context=ctx.obj) as rossum:
        schemas = rossum.get_schemas((QUEUES,))

    table = [
        [schema["id"], schema["name"], ", ".join(str(s.get("id", "")) for s in schema["queues"])]
        for schema in schemas
    ]

    click.echo(tabulate(table, headers=["id", "name", "queues"]))



#!/usr/bin/env python3
import warnings
from copy import deepcopy
from typing import List, Callable, Optional, Tuple, Set

import click as click

from rossum.lib import DataPointDict, DataPoints


def traverse_datapoints(
    datapoints: List[dict], transformation: Callable, parent_categories: List[str] = None, **kwargs
) -> List[dict]:
    dummy_root = {"category": "root", "id": None, "children": datapoints}
    return _traverse_datapoints([dummy_root], transformation, parent_categories, **kwargs)[0][
        "children"
    ]


def _traverse_datapoints(
    datapoints: List[dict], transformation: Callable, parent_categories: List[str] = None, **kwargs
) -> List[dict]:
    new_datapoints = []
    parent_categories = parent_categories or []
    for datapoint in datapoints:
        new_datapoint = deepcopy(datapoint)
        category = datapoint["category"]
        if category != "datapoint":
            parent_categories_ = parent_categories[:] + [category]
            children = new_datapoint.pop("children", [])
            if category == "multivalue" and children:
                [new_datapoint["children"]] = _traverse_datapoints(
                    [children], transformation, parent_categories_, **kwargs
                )
            elif category == "multivalue":
                new_datapoint["children"] = None
            elif category in ("tuple", "section", "root"):
                new_datapoint["children"] = _traverse_datapoints(
                    children, transformation, parent_categories_, **kwargs
                )
        new_datapoint = transformation(new_datapoint, parent_categories, **kwargs)
        if new_datapoint:
            new_datapoints.append(new_datapoint)

    return new_datapoints


def substitute_options(
    datapoint: DataPointDict, parent_categories: List[str], id_: str, options: List[dict]
) -> DataPointDict:
    if (
        datapoint["category"] != "datapoint"
        or datapoint["type"] != "enum"
        or datapoint["id"] != id_
    ):
        return datapoint

    new_datapoint = deepcopy(datapoint)
    return {**new_datapoint, "options": options}


def remove(
    datapoint: DataPointDict, parent_categories: List[str], ids: Tuple[str, ...]
) -> Optional[DataPointDict]:
    if datapoint["id"] in ids:
        if parent_categories and "multivalue" == parent_categories[-1]:
            warnings.warn("Cannot delete child of a multivalue.")
            return datapoint
        else:
            return None
    else:
        return datapoint


def wrap_in_multivalue(
    datapoint: DataPointDict, parent_categories: List[str], exclude_ids: Set[str]
) -> DataPointDict:
    if (
        "multivalue" in parent_categories
        or datapoint["category"] in ("multivalue", "section", "root")
        or datapoint["id"] in exclude_ids
    ):
        return datapoint
    return {
        "id": f'{datapoint["id"]}_multi',
        "label": datapoint["label"],
        "children": {"use_rir_content": True, **datapoint},
        "category": "multivalue",
        "max_occurrences": None,
        "min_occurrences": None,
    }


def add(
    datapoint: dict,
    parent_categories: List[str],
    parent_id: str,
    datapoint_to_add: DataPointDict,
    place_before: Optional[str] = None,
) -> dict:
    if datapoint["id"] != parent_id:
        return datapoint

    if datapoint["category"] in ("tuple", "section", "root"):
        new_datapoint = deepcopy(datapoint)
        if place_before is None:
            new_datapoint["children"].append(datapoint_to_add)
        else:
            new_datapoint["children"].insert(
                _find_index_of_id(place_before, new_datapoint["children"]), datapoint_to_add
            )
        return new_datapoint
    elif datapoint["category"] == "multivalue" and not datapoint.get("children"):
        new_datapoint = deepcopy(datapoint)
        new_datapoint["children"] = datapoint_to_add
        return new_datapoint
    elif datapoint["category"] == "multivalue":
        warnings.warn("Cannot add child to a filled multivalue.")
        return datapoint
    else:
        return datapoint


def _find_index_of_id(id_: str, children: DataPoints) -> int:
    for i, child in enumerate(children):
        if child["id"] == id_:
            return i
    else:
        raise click.ClickException(f"Not found ID '{id_}' to place behind.")


def change(
    datapoint: DataPointDict,
    parent_categories: List[str],
    id_: str,
    to_change: DataPointDict,
    filtered_categories: Tuple[str],
) -> DataPointDict:
    is_of_id = id_ in (datapoint["id"], "ALL")
    is_of_category = not filtered_categories or datapoint["category"] in filtered_categories
    if not (is_of_id and is_of_category):
        return datapoint
    else:
        return {**datapoint, **to_change}


def get(datapoints: List[dict], id_: str) -> Optional[dict]:
    res = None

    for datapoint in datapoints:
        if datapoint["id"] == id_:
            return datapoint

        category = datapoint["category"]
        if category in ("section", "tuple"):
            res = get(datapoint["children"], id_)
        elif category == "multivalue":
            res = get([datapoint["children"]], id_)
        if res is not None:
            return res
    return res


def _new_datapoint(datapoint_to_add: DataPointDict) -> DataPointDict:
    try:
        id_ = datapoint_to_add.pop("id")
    except KeyError as e:
        raise click.BadArgumentUsage("Missing key 'id'.") from e

    category = datapoint_to_add.pop("category", "datapoint")
    default = {"id": id_, "label": id_}
    if category == "datapoint":
        default = {**default, "rir_field_names": [], **_new_singlevalue(datapoint_to_add)}
    elif category == "multivalue":
        default = {
            **default,
            "children": None,
            "default_value": None,
            "min_occurrences": None,
            "max_occurrences": None,
        }
    elif category == "tuple":
        default = {**default, "rir_field_names": [], "children": []}
    elif category == "section":
        raise click.BadArgumentUsage("Cannot add section.")
    else:
        raise click.BadArgumentUsage("Unknown category.")
    return {**default, **datapoint_to_add, "category": category}


def _new_singlevalue(datapoint_to_add: DataPointDict) -> DataPointDict:
    type_ = datapoint_to_add.pop("type", "string")
    default: DataPointDict = {
        "width_chars": 10,
        "default_value": None,
        "constraints": {"required": False},
    }
    if type_ == "enum":
        default = {**default, "options": [{"value": "0", "label": "---"}]}
    elif type_ == "date":
        default = {**default, "format": "D. M. YYYY"}
    elif type_ == "number":
        default = {**default, "format": "# ##0.#"}
    elif type_ != "string":
        raise click.BadArgumentUsage("Unknown type.")

    return {**default, "type": type_}



#!/usr/bin/env python3
from setuptools import setup, find_packages

setup(
    name="rossum",
    version="3.5.0",
    description="Command line interface for controlling the Rossum platform",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://developers.rossum.ai/",
    author="Rossum developers",
    author_email="support@rossum.ai",
    license="MIT",
    project_urls={
        "Source": "https://github.com/rossumai/rossum",
        "Tracker": "https://github.com/rossumai/rossum/issues",
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages(exclude=("tests*",)),
    install_requires=[
        "pandas",
        "click",
        "click-shell",
        'xlrd > 1.2.0;python_version>"3.6"',
        'xlrd==1.2.0;python_version=="3.6"',
        "requests",
        "jsondiff",
        "tabulate",
        'dataclasses;python_version<"3.7"',
        "openpyxl>=2.6",
        "jmespath",
        "polling2",
        "more_itertools",
    ],
    python_requires=">=3.6",
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "pytest-cov", "requests_mock", "pytest-click"],
    zip_safe=False,
    entry_points={"console_scripts": ["rossum = rossum.main:entry_point"]},
)



import click
from click_shell import shell

import rossum.schema.commands
from rossum import (
    tools,
    configure,
    user,
    workspace,
    queue,
    connector,
    hook,
    document,
    __version__,
    CTX_DEFAULT_PROFILE,
    CTX_PROFILE,
    user_assignment,
    password,
)


@shell(
    prompt="rossum> ",
    intro="Welcome to the rossum interactive mode. Start with `help` and `configure`.",
    context_settings={"help_option_names": ["-h", "--help"]},
)
@click.version_option(__version__)
@click.option(
    "-p",
    "--profile",
    default=CTX_DEFAULT_PROFILE,
    type=str,
    help="Profile name.",
    show_default=True,
)
@click.pass_context
def entry_point(ctx: click.Context, profile: str) -> None:
    ctx.obj = {CTX_PROFILE: profile}


entry_point.add_command(tools.cli)
entry_point.add_command(rossum.schema.commands.cli)
entry_point.add_command(user.cli)
entry_point.add_command(user_assignment.cli)
entry_point.add_command(workspace.cli)
entry_point.add_command(queue.cli)
entry_point.add_command(document.cli)
entry_point.add_command(connector.cli)
entry_point.add_command(hook.cli)
entry_point.add_command(configure.cli)
entry_point.add_command(password.cli)
